from __future__ import unicode_literals

from glyphs.rw.RWGlyph import RWGlyph
from glyphs.utils.StringUtils import StringUtils


class ResettableGlyph(RWGlyph):
    """
        Special read/write glyph is designed to read data from an object and write data back.

        It also has the ability to write a value expressing a "force reset" (when supported) into the object.
    """

    def __init__(self, reset_w_path, reset_value,
                 r_path, w_path, r_types=None,
                 w_types=None,
                 reset_w_type=None,
                 r_translation_function=None,
                 w_translation_function=StringUtils.to_unicode,
                 r_default_value=None,
                 w_allow_none=False,
                 ):
        """
            Initializer for a resettable R/W glyph instance.

            @param reset_w_path: The "path" pointing to where the data needs to be written to express a "force
            reset" (when supported).
            For multilevel keys (nest level):
            - if the path is a unicode, use L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of keys.
            - if the path is a collection, each level is a unicode in the collection.
            @type reset_w_path: UnicodeType or TupleType
            @param reset_w_type: (Optional) A string describing the type of data expected to be found once the data
            is written. For each matching level in L{reset_w_path}, a key-value pair may be specified, not all
            levels are required to have a type defined.
            - if L{reset_w_type} is a unicode, use L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of types. To omit a level of type, do not specify a key-value pair.
            - if L{reset_w_type} is a collection, each level of types is a unicode in the collection. To omit a
            level of type, place C{None} at the skipped level.

            The level of types are key-value pairs. The key defines where the type information can be found at
            a given level in the source. The values specifies the type expected. The key and value separator
            <glyphs.api.ROGlyph.ROGlyph.KEY_VALUE_SEPARATOR>}.
            There is no need to specify every level of type to match L{reset_w_path},
            as any long-tail levels that are not specified will be interpreted as having no key-value
            pairs.


            @param reset_value: The reset value

            @precondition: isinstance(reset_w_path, unicode) or all(isinstance(u, unicode) for u in reset_w_path)
            @precondition: len(reset_w_path) > 0
            @precondition: reset_w_type is None or isinstance(reset_w_type, (unicode, tuple))
            @precondition: reset_w_type is None or len(reset_w_type) > 0
            @precondition: reset_w_type is None or isinstance(reset_w_type, tuple) or (
                                                                                     (
                                                                                     isinstance(reset_w_path, unicode)
                                                                                     and reset_w_type.count(ROGlyph.NAME_SPACE_SEPARATOR) <= reset_w_path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                                     )
                                                                                     or reset_w_type.count(ROGlyph.NAME_SPACE_SEPARATOR) <= len(reset_w_path)
                                                                                     )
            @precondition: reset_w_type is None or isinstance(reset_w_type, unicode) or (
                                                                                     (
                                                                                     isinstance(reset_w_path, unicode)
                                                                                     and len(reset_w_type) <= reset_w_path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                                     )
                                                                                     or len(reset_w_type) <= len(reset_w_path)
                                                                                     )
            @precondition: reset_w_type is None or len(reset_w_type) > 0

            @see: glyphs.rw.RWGlyph.RWGlyph.__init__
        """
        super(ResettableGlyph, self).__init__(
                                              r_path,
                                              w_path,
                                              r_types,
                                              w_types,
                                              r_translation_function,
                                              w_translation_function,
                                              r_default_value,
                                              w_allow_none,
                                              )

        if reset_w_type is None:
            reset_w_type = w_types

        self.__dict__["__reset_w_path_type"] = self._generate_path_type_paired_sequence(reset_w_path, reset_w_type,)

        self.__dict__["__reset_value"] = reset_value

    @property
    def iter_reset_w_path_type(self):
        """
            Returns a new iterator through the configured sequence of triplet of the sub path, their matching
            types as a pair and a boolean expressing whether they are the last triplet in the sequence.

            The iterator makes up a typed path to reset write data into an object.

            @rtype: collections.Iterable

            @postcondition: next(return, None) is not None
            @postcondition: all(
                                (
                                    isinstance(x, tuple)
                                    and len(x) == 3
                                    and isinstance(x[0], bool)
                                    and isinstance(x[1], unicode) and len(x[1]) > 0
                                    and (
                                        x[2] is None
                                        or (
                                            isinstance(x[2], tuple) and len(x[2]) == 2
                                            and all(isinstance(v) for v in x[2])
                                           )
                                        )
                                )
                                for x in return
                                )
        """
        return iter(self.__dict__["__reset_w_path_type"])

    @property
    def reset_value(self):
        """
            Gets the value indicating a reset for this entity.
        """
        return self.__dict__["__reset_value"]



from __future__ import unicode_literals

import collections
from glyphs.helpers.ImmutableType import ImmutableType
from glyphs.ro.ROGlyph import ROGlyph
from glyphs.rw.RWGlyph import RWGlyph
from glyphs.rw.ResettableGlyph import ResettableGlyph

import six


class DictUtils(six.with_metaclass(ImmutableType)):
    """
        Utility methods for working with glyphs and dictionaries.
    """

    @staticmethod
    def get(source, glyph, force_default_to_none=False):
        """
            Returns the value stored in the given L{source} if it has the B{full} L{path and
            types<glyphs.api.ROGlyph.ROGlyph.iter_r_path_type>} held by L{glyph}.

            Otherwise:
            - if any intermediary pieces of the path is not in L{source} raises a C{KeyError}
            - if only the last piece of the path is missing returns L{the default value<glyphs.api.ROGlyph.
            ROGlyph.default_return>}.

            Any returned value is passed through the L{translation function of the glyph<glyphs.api.ROGlyph.
            ROGlyph.r_translation_function>} if one exists.

            If a call to L{DictUtils.in_} with the same L{source} and L{glyph} returns C{False}, this method
            will throw an error.
            If a call to L{DictUtils.in_} with the same L{source} and L{glyph} returns C{True}, this method
            will not throw an error and return a value (defaulted or translated)

            @type source: collections.Mapping
            @type glyph: ROGlyph
            @param force_default_to_none: If C{True}, L{default_return} will be returned if the 'raw' value
            found is C{None} (checked before translation function is applied).

            @raise KeyError: if any intermediary pieces of the path is not in L{source}
            @raise TypeError: if any intermediary pieces of the path does not match the expected type found in L{source}
        """
        Mapping = collections.Mapping
        assert isinstance(source, Mapping)
        assert isinstance(glyph, ROGlyph)

        # other preconditions tested below

        current_dict = source
        Container = collections.Container
        default_return = glyph.r_default_value

        for sub_path, source_type in glyph.iter_r_path_type:

            if not isinstance(current_dict, Mapping):
                raise KeyError('Could not find {} in the given dictionary'.format(sub_path))

            if isinstance(source_type, tuple):
                key = source_type[0]

                if (key not in current_dict
                    or isinstance(current_dict[key], Container) # saving the serialization cost as it is not going to work
                    or source_type[1] != unicode(current_dict[key])):
                    raise TypeError('Type mismatch for {} in the given dictionary'.format(sub_path))
            else:
                assert source_type is None

            # Cannot be current_dict.get() because not all collections.Mapping have get().
            current_dict = current_dict[sub_path] if sub_path in current_dict else default_return

        if (
            current_dict == default_return # type could be different in the case of string vs unicode.
            or (force_default_to_none and current_dict is None)
            ):
            return default_return

        t = glyph.r_translation_function
        if t and current_dict is not None:
            return t(current_dict)

        return current_dict

    @staticmethod
    def in_(source, glyph):
        """
            Returns  C{True} if the given L{source} has the B{full} L{path and types<glyphs.api.ROGlyph.
            ROGlyph.iter_r_path_type>} held by L{glyph}. Otherwise, returns C{False}.

            @type source: collections.Mapping
            @type glyph: ROGlyph

            @rtype: BooleanType
        """
        Mapping = collections.Mapping
        assert isinstance(source, Mapping)

        assert isinstance(glyph, ROGlyph)
        # other preconditions tested below

        current_dict = source
        Container = collections.Container
        iter_r_path_type = glyph.iter_r_path_type
        is_last = False

        for is_last, sub_path, source_type in iter_r_path_type:
            if isinstance(source_type, tuple):
                assert len(source_type) == 2

                key = source_type[0]

                if (
                    key not in current_dict
                    or isinstance(current_dict[key], Container) # saving the serialization cost as it is not going to work
                    or source_type[1] != unicode(current_dict[key])
                    ):
                    return False
            else:
                assert source_type is None

            if sub_path in current_dict:
                current_dict = current_dict[sub_path]
            else:
                return False

        return is_last

    @staticmethod
    def set(destination, glyph, value):
        """
            Sets the given L{value} in the L{destination} using L{utils.api.APIGlyph.APIGlyph.
            target_source_names}.

            The value is translated using L{utils.api.APIGlyph.APIGlyph.target_translation_func} if any is found. It is
            set 'raw' otherwise.

            If the L{value} or the translated value is C{None} then this method is a NOOP.

            @type destination: collections.MutableMapping
            @type glyph: APIGlyph

        """
        assert isinstance(glyph, RWGlyph)

        if glyph.w_translation_function:
            value = glyph.w_translation_function(value)

        if value is not None or glyph.w_allow_none:
            DictUtils.__set(destination, glyph.iter_w_path_type, value)

    @staticmethod
    def set_reset_value(destination, glyph):
        """
            Sets the L{reset value of the glyph<utils.api.AbstractResettableGlyph
            .AbstractResettableGlyph.reset_value>}, which will be interpreted by the API as a
            'reset', in the L{destination} using L{the reset source names of the glyph<utils.api
            .AbstractResettableGlyph.AbstractResettableGlyph.reset_target_source_names>} and
            the L{target type names<utils.api.APIGlyph.APIGlyph.target_type_names>}.

            @type destination: collections.MutableMapping
            @type glyph: AbstractResettableGlyph
        """
        assert isinstance(glyph, ResettableGlyph)
        # other preconditions tested in called func

        # when resetting, the reset value may not always be at the same place as the target type,
        # so we have to allow the depth mismatch
        DictUtils.__set(destination, glyph.iter_reset_w_path_type, glyph.reset_value,)

    @staticmethod
    def __set(destination, w_path_type, value,):
        """
            Sets the given L{value} in the L{destination} using L{target_source_names}.

            It recursively builds the sub dictionaries if the number of source names is greater than
            one. Each name is then used as L{depth} in the L{destination}. When on the last element
            of L{target_source_names} the key and value from L{root_type_key_values} is set and the
            method returns.

            @type destination: collections.MutableMapping
            @param root_type_key_values: Describes the key and value, or lack thereof, of all levels.
            @type root_type_key_values: tuple
            @type target_source_names: tuple
            @type depth: int

            @precondition: next(w_path_type, None,) is not None
        """
        assert isinstance(destination, collections.MutableMapping), type(destination)
        # other preconditions tested below

        is_last, sub_path, type_tuple, = next(w_path_type,)

        if is_last is False:
            sub_dict = destination.get(sub_path)
            if sub_dict is None:
                sub_dict = {}
                destination[sub_path] = sub_dict
                if type_tuple is not None:
                    key, value, = type_tuple

                    if key not in destination:
                        destination[key] = value
                    else:
                        # dev check
                        assert destination[key] == value

            DictUtils.__set(sub_dict, w_path_type, value,)
            return

        if type_tuple is not None:
            key, value, = type_tuple

            if key not in destination:
                destination[key] = value
            else:
                # dev check
                assert destination[key] == value

        destination[sub_path] = value

    __slots__ = tuple()



from __future__ import unicode_literals

from glyphs.ro.ROGlyph import ROGlyph
from glyphs.utils.StringUtils import StringUtils


class RWGlyph(ROGlyph):
    """
        Read/write glyph is designed to read data from an object and write data back.
    """

    def __init__(self, r_path, w_path, r_types=None, w_types=None,
                 r_translation_function=None,
                 w_translation_function=StringUtils.to_unicode,
                 r_default_value=None,
                 w_allow_none=False,
                 ):
        """
            Initializer for a R/W glyph instance.

            @param w_path: The "path" pointing to where the data needs to be written.
            For multilevel keys (nest level):
            - if the path is a unicode, use L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of keys.
            - if the path is a collection, each level is a unicode in the collection.
            @type w_path: unicode or collections.Sequence
            @param w_types: (Optional) A string describing the type of data expected to be found once the data
            is written. For each matching level in L{w_path}, a key-value pair may be specified, not all
            levels are required to have a type defined.
            - if L{w_types} is a unicode, use L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of types. To omit a level of type, do not specify a key-value pair.
            - if L{w_types} is a collection, each level of types is a unicode in the collection. To omit a
            level of type, place C{None} at the skipped level.

            The level of types are key-value pairs. The key defines where the type information can be found at
            a given level in the source. The values specifies the type expected. The key and value separator
            <glyphs.api.ROGlyph.ROGlyph.KEY_VALUE_SEPARATOR>}.
            There is no need to specify every level of type to match L{w_path},
            as any long-tail levels that are not specified will be interpreted as having no key-value
            pairs.
            @param w_translation_function: function used to translate the data to be written at L{w_path}
            to any target format.
            This is meant to be a very simple function for pure translate / transtyping. No intelligence here.
            If C{None}, then no translation is needed and the data should be passed through.
            @param w_allow_none: C{True} if the object supports a C{None} value (when written into). Otherwise,
            C{False}.

            @precondition: isinstance(w_path, unicode) or all(isinstance(u, unicode) for u in w_path)
            @precondition: len(w_path) > 0
            @precondition: w_types is None or isinstance(w_types, (unicode, tuple))
            @precondition: w_types is None or len(w_types) > 0
            @precondition: w_types is None or isinstance(w_types, tuple) or (
                                                                                     (
                                                                                     isinstance(w_path, unicode)
                                                                                     and w_types.count(ROGlyph.NAME_SPACE_SEPARATOR) <= w_path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                                     )
                                                                                     or w_types.count(ROGlyph.NAME_SPACE_SEPARATOR) <= len(w_path)
                                                                                     )
            @precondition: w_types is None or isinstance(w_types, unicode) or (
                                                                                     (
                                                                                     isinstance(w_path, unicode)
                                                                                     and len(w_types) <= w_path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                                     )
                                                                                     or len(w_types) <= len(w_path)
                                                                                     )
            @precondition: w_types is None or len(w_types) > 0
            @precondition: w_translation_function is None or callable(w_translation_function)

            @see: glyphs.api.ROGlyph.ROGlyph.__init__
        """
        super(RWGlyph, self).__init__(
                                      r_path,
                                      r_types,
                                      r_translation_function,
                                      r_default_value,
                                      )

        self.__dict__["__w_translation_function"] = w_translation_function
        self.__dict__["__w_path_type"] = self._generate_path_type_paired_sequence(w_path, w_types,)
        self.__dict__["__w_allow_none"] = w_allow_none

    @property
    def iter_w_path_type(self):
        """
            Returns a new iterator through the configured sequence of triplet of the sub path, their matching
            types as a pair and a boolean expressing whether they are the last triplet in the sequence.

            The iterator makes up a typed path to write data into an object.

            @rtype: collections.Iterable

            @postcondition: next(return, None) is not None
            @postcondition: all(
                                (
                                    isinstance(x, tuple)
                                    and len(x) == 3
                                    and isinstance(x[0], bool)
                                    and isinstance(x[1], unicode) and len(x[1]) > 0
                                    and (
                                        x[2] is None
                                        or (
                                            isinstance(x[2], tuple) and len(x[2]) == 2
                                            and all(isinstance(v) for v in x[2])
                                           )
                                        )
                                )
                                for x in return
                                )
        """
        return iter(self.__dict__["__w_path_type"])

    @property
    def w_translation_function(self):
        """
            Returns a function used to translate a value to the format expected in the object.

            Note: this is meant to be a very simple function for pure translate / transtyping. No
            intelligence here.

            If C{None}, then no translation is needed and the data should be passed through.

            @postcondition: return is None or callable(return)

        """
        return self.__dict__["__w_translation_function"]

    @property
    def w_allow_none(self):
        """
            Returns C{True} if the object supports s C{None} value (when written into). Otherwise, returns
            C{False}.

            @rtype: bool
        """
        return self.__dict__["__w_allow_none"]



from __future__ import unicode_literals

import collections
import itertools

from glyphs.helpers.ImmutableObject import ImmutableObject


class ROGlyph(ImmutableObject):
    """
        R/O glyph designed to read data from an object (source of the data).
    """

    NAME_SPACE_SEPARATOR = '>'
    """
        Token used to separate the 'levels' for the keys-value pairs.
    """

    KEY_VALUE_SEPARATOR = ':'
    """
        Token used to separate the key and values for the key-value pairs.
    """

    def __init__(self, r_path, r_types=None, r_translation_function=None, r_default_value=None):
        """
            Initializer for a R/O glyph.

            @param r_path: The "path" pointing to the source of data.
            For multilevel keys (nest level):
            - if the path is a unicode, use L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of keys.
            - if the path is a collection, each level is a unicode in the collection.
            @type r_path: unicode or collections.Sequence
            @param r_types: (Optional) A string describing the type of data expected to be found in the
            source data. For each matching level in L{r_path}, a key-value pair may be specified, not all
            levels are required to have a type defined.
            - if L{r_types} is a unicode, use L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of types. To omit a level of type, do not specify a key-value pair.
            - if L{r_types} is a collection, each level of types is a unicode in the collection. To omit a
            level of type, place C{None} at the skipped level.

            The level of types are key-value pairs. The key defines where the type information can be found at
            a given level in the source. The values specifies the type expected. The key and value separator
            <glyphs.api.ROGlyph.ROGlyph.KEY_VALUE_SEPARATOR>}.
            There is no need to specify every level of type to match L{r_path},
            as any long-tail levels that are not specified will be interpreted as having no key-value
            pairs.
            @param r_translation_function: function used to translate the data found at L{r_path}
            to any target format.
            This is meant to be a very simple function for pure translate / transtyping. No intelligence here.
            If C{None}, then no translation is needed and the data should be passed through.
            @param r_default_value: value to use as a default value. To be used when no data is found at
            L{r_path}.

            @precondition: isinstance(r_path, unicode) or all(isinstance(u, unicode) for u in r_path)
            @precondition: len(r_path) > 0
            @precondition: r_types is None or isinstance(r_types, (unicode, tuple))
            @precondition: r_types is None or len(r_types) > 0
            @precondition: r_types is None or isinstance(r_types, tuple) or (
                                                                                     (
                                                                                     isinstance(r_path, unicode)
                                                                                     and r_types.count(ROGlyph.NAME_SPACE_SEPARATOR) <= r_path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                                     )
                                                                                     or r_types.count(ROGlyph.NAME_SPACE_SEPARATOR) <= len(r_path)
                                                                                     )
            @precondition: r_types is None or isinstance(r_types, unicode) or (
                                                                                     (
                                                                                     isinstance(r_path, unicode)
                                                                                     and len(r_types) <= r_path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                                     )
                                                                                     or len(r_types) <= len(r_path)
                                                                                     )
            @precondition: r_types is None or len(r_types) > 0
            @precondition: r_translation_function is None or callable(r_translation_function)
        """
        assert r_translation_function is None or callable(r_translation_function)

        self.__dict__["__r_path_type"] = self._generate_path_type_paired_sequence(r_path, r_types,)

        self.__dict__["__r_translation_function"] = r_translation_function
        self.__dict__["__r_default_value"] = r_default_value

    def __repr__(self):
        return "{}({},)".format(self.__class__.name, self.r_path,)

    @property
    def iter_r_path_type(self):
        """
            Returns a new iterator through the configured sequence of triplet of the sub path, their matching
            types as a pair and a boolean expressing whether they are the last triplet in the sequence.

            The iterator makes up a typed path to read data out of an object.

            @rtype: collections.Iterable

            @postcondition: next(return, None) is not None
            @postcondition: all(
                                (
                                    isinstance(x, tuple)
                                    and len(x) == 3
                                    and isinstance(x[0], bool)
                                    and isinstance(x[1], unicode) and len(x[1]) > 0
                                    and (
                                        x[2] is None
                                        or (
                                            isinstance(x[2], tuple) and len(x[2]) == 2
                                            and all(isinstance(v) for v in x[2])
                                           )
                                        )
                                )
                                for x in return
                                )
        """
        return iter(self.__dict__["__r_path_type"])

    @property
    def r_default_value(self):
        """
            Returns a value to use in case no suitable value is found in the object.
        """
        return self.__dict__["__r_default_value"]

    @property
    def r_translation_function(self):
        """
            Returns a function used to translate the value found in the object to a target
            format.

            Note: this is meant to be a very simple function for pure translate / transtyping. No
            intelligence here.

            If C{None}, then no translation is needed and the data should be passed through.

            @postcondition: return is None or callable(return)
        """
        return self.__dict__["__r_translation_function"]

    def _generate_path_type_paired_sequence(self, path, types,):
        """
            Returns a sequence of triplet of the sub path, their matching L{types} as a pair and a boolean expressing whether
            they are the last triplet in the sequence.

            @param path: The "path" to walk to reach the data.
            For multilevel keys (nest level):
            - if the path is a unicode, uses L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of keys.
            - if the path is a collection, each level is a unicode in the collection.
            @type path: unicode or collections.Sequence
            @param types: (Optional) A string describing the type of data expected to be found in the
            source data. For each matching level in L{path} (sub path), a key-value pair may be specified, not
            all levels are required to have a type defined.
            - if L{types} is a unicode, uses L{a name space separator<glyphs.api.ROGlyph.ROGlyph.NAME_SPACE_SEPARATOR>}
            between the various level of types. To omit a level of type, do not specify a key-value pair at
            that level.
            - if L{types} is a collection, each level of types is a unicode in the collection. To omit a
            level of type, place C{None} at the skipped level.

            The level of types are key-value pairs. The key defines where the type information can be found at
            a given level in the source. The values specifies the type expected. The key and value separator
            <glyphs.api.ROGlyph.ROGlyph.KEY_VALUE_SEPARATOR>}.
            There is no need to specify every level of type to match L{path},
            as any long-tail levels that are not specified will be interpreted as having no key-value
            pairs.

            e.g.
            if the separator is a period '.' as a name space separator and ':' as a key separator to express
            and object which is a series of nested dictionaries:
            path: sub_path1.sub_path2.sub_path3
            value: 1
            types: ..xsi:Entity
            object: {'sub_path1': {'sub_path2': {'sub_path3': 1, 'xsi':'Entity' } } }

            @rtype: collections.Sequence

            @precondition: types is None or isinstance(types, tuple) or (
                                                                             (
                                                                             isinstance(path, unicode)
                                                                             and types.count(ROGlyph.NAME_SPACE_SEPARATOR) <= path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                             )
                                                                             or types.count(ROGlyph.NAME_SPACE_SEPARATOR) <= len(path)
                                                                        )
            @precondition: types is None or isinstance(types, unicode) or (
                                                                             (
                                                                             isinstance(path, unicode)
                                                                             and len(types) <= path.count(ROGlyph.NAME_SPACE_SEPARATOR)
                                                                             )
                                                                             or len(types) <= len(path)
                                                                          )
            @precondition: types is None or len(types) > 0

            @postcondition: len(return) > 0
            @postcondition: all(
                                    (
                                        isinstance(x, tuple)
                                        and len(x) == 3
                                        and isinstance(x[0], bool)
                                        and isinstance(x[1], unicode) and len(x[1]) > 0
                                        and (
                                            x[2] is None
                                            or (
                                                isinstance(x[2], tuple) and len(x[2]) == 2
                                                and all(isinstance(v) for v in x[2])
                                               )
                                            )
                                    )
                                    for x in return
                                )
        """

        if isinstance(path, unicode):
            path_tuple = tuple(path.split(ROGlyph.NAME_SPACE_SEPARATOR))
        else:
            assert isinstance(path, collections.Sequence) # pre
            path_tuple = tuple(path)
            assert all(isinstance(u, unicode) for u in path_tuple) # pre

        assert path_tuple

        max_index = len(path_tuple) - 1


        if types is None:
            return None

        if isinstance(types, unicode):
            types_tuple = tuple(None if x == '' else x for x in types.split(ROGlyph.NAME_SPACE_SEPARATOR))
        else:
            assert isinstance(types, tuple)
            assert all(l is None or isinstance(l, unicode) for l in types)
            types_tuple = types

        # If this glyph is created with types_tuple, but they're all None (or we got a tuple of Nones)
        # let's revert it from the tuple back to None, to save cycles when the glyph is used
        # for setting values. This would happen if someone gave '..' as the types_tuple
        if all(t is None for t in types_tuple):
            return None

        assert types_tuple

        types_tuple = tuple(None if x is None else tuple(x.split(ROGlyph.KEY_VALUE_SEPARATOR)) for x in types_tuple)

        if len(types_tuple) < len(path_tuple):
            # pad anything that has been left unspecified.
            types_tuple += (None,) * (len(path_tuple) - len(types_tuple))
        else:
            assert len(types_tuple) == len(path_tuple) # pre

        assert len(types_tuple) == len(path_tuple) # dev check

        return tuple(
                      (idx == max_index, sub_path, type_name,)
                      for idx, (sub_path, type_name,) in enumerate(
                                                                  itertools.izip(
                                                                                 path_tuple,
                                                                                 types_tuple,
                                                                                 )
                                                                   )
                      )



from __future__ import unicode_literals

import six

from glyphs.helpers.ImmutableType import ImmutableType


class ImmutableObject(six.with_metaclass(ImmutableType)):
    """
        Abstract class making any implementation and instance of this class immutable.

        i.e. no instance and class attributes will able to be set.
    """

    def __setattr__(self, name, value):
        assert False, "No sets allowed for instances of %(class_name)s" % {"class_name" :self.__class__.__name__}

    __slots__ = tuple()



from __future__ import unicode_literals

class ImmutableType(type):
    """
        Abstract type making any implementing class immutable.

        i.e. no static attributes will able to be set.
    """

    def __setattr__(cls, name, value):
        assert False, "No sets allowed for %(class_name)s" % {"class_name" :cls.__name__}



#!/usr/bin/env python

from distutils.core import setup
from setuptools import find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
        name='glyphs',
        version='0.1',
        description='Swiss army knife of data extraction',
        long_description=long_description,
        long_description_content_type="text/markdown",
        author='slorg1',
        url='https://github.com/slorg1/glyphs',
        packages=[
                  'glyphs',
                  'glyphs.helpers',
                  'glyphs.ro',
                  'glyphs.rw',
                  'glyphs.utils',
                  ],
        package_dir={'':'src'},
        install_requires=[
              "six >= 1.10.0",
          ],
        classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Intended Audience :: Developers',
            'Topic :: Software Development :: Build Tools',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
          ],
     )



from __future__ import unicode_literals


class StringUtils(object):
    """
        Utility class for handling strings.
    """

    @staticmethod
    def to_unicode(value):
        """
            Returns the string representation of the given L{value} if it is not C{None}. Otherwise,
            returns C{None}.

            @postcondition: (value is None) == (return is None)
            @postcondition: return is None or isinstance(return, unicode)
        """
        return value if value is None else unicode(value)

    __slots__ = tuple()
